/*
 * Lua Commander - Lua Debugger
 *
 * Copyright (C) 1999-2000, Antonio E. Scuri, Tecgraf/PUC-Rio.
 */


#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include <iup.h>
#include <iupkey.h>
#include <lua.h>
#include <luadebug.h>

#include "lconsole.h"
#include "ldebug.h"
#include "utl.h"


enum {LINEDEF_CFUNC = -1, LINEDEF_MAIN = 0};
enum {DEBUG_INACTIVE, DEBUG_ACTIVE, DEBUG_RUN, DEBUG_STEP, DEBUG_PAUSED, DEBUG_STOPPED};
enum {MODE_LINES, MODE_OUT};
enum {STEP_INTO, STEP_OVER};
enum {FUNC_INSIDE, FUNC_OUTSIDE};

#define MAX_BREAKS 10240


/* instead of a list of breakpoints at a line of a file, 
   we store a list of lines that has breakpoits */
typedef struct _luaDebugBreak
{
  char* filename[100];  /* copy of the file name that has a break on this line */
  int count;            /* number of breakpoints on this line */
}luaDebugBreak;


static struct _luaDebug
{
  Ihandle *iupDialog,
          *btStop, *btPause, *btPlay, *btStep, 
          *btCurLine, *btCurLineBreak, 
          *btSetLocal, *btPrintLocal,
          *btPrintLevel, *btPrintStack,
          *tgActive,
          *lsStack,
          *lsLocal,
          *lbSource, *mlSource;
  luaDebugBreak breaks[MAX_BREAKS];
  void (*old_error_func)(void); /* error function of the Lua console */
  char *last_filename;  /* a pointer to the last filename into the source multiline, used to smart update the source multiline */
  char *filename;       /* a copy of the current filename beeing debugged, used to find breakpoints */
  int close,            /* return value of the Close callback */
      state,            /* debug state, check the enum above */
      wait_time,        /* animation delay time in miliseconds */
      levels_to_stop,   /* number of levels in the stack when a stop occurs, decreased in the hooks until stops completly. */
      last_stack_update_level, /* last stack level that I update the interface stack list, so I know if do not need to update again */
      total_breaks,     /* total number of breakpoints */
      step_mode, step_func_mode, step_lines_total, /* step behavior, changed only by the interface controls */
      step_lines_count, /* number of lines executed in a STEP, when mode is LINES */
      step_func_state,  /* informs if I'm in or out a function call after I set step OVER or OUT */
      step_func_level;  /* store the level in the stack where you gets in, so you know when you are out */
}debug;


/*************************************************************************/
/*                         Breaks                                        */
/*************************************************************************/

static void DebugUpdateSourceBreak(int line);

static void DebugRemoveBreak(int num)
{
  int i,j, n=0;
  for(i=0; i < MAX_BREAKS; i++)
  {
    luaDebugBreak* break_line = &debug.breaks[i];

    for (j=0; j < break_line->count; j++)
    {
      if (n == num)
      {
        int update = 0;
        if (strcmp(debug.filename, break_line->filename[j]) == 0)
          update = 1;

        free(break_line->filename[j]);

        if (break_line->count == 1)
          break_line->filename[j] = NULL;
        else
          break_line->filename[j] = break_line->filename[break_line->count-1];

        break_line->count--;
        debug.total_breaks--;

        if (update)
          DebugUpdateSourceBreak(i);

        return;
      }

      n++;
    }
  }
}

static void DebugBreakFillList(char* list[])
{
  int i,j, n=0;
  char buffer[200];
  for(i=0; i < MAX_BREAKS; i++)
  {
    luaDebugBreak* break_line = &debug.breaks[i];

    for (j=0; j < break_line->count; j++)
    {
      char* title = utlGetFileTitle(break_line->filename[j]);
      sprintf(buffer, "Line %03d of %s", i, title);
      free(title);

      list[n] = utlStrDup(buffer);
      n++;

      if (n == debug.total_breaks)
        return;
    }
  }
}

static int DebugRemoveBreakDlg(void)
{
  int ret, i;
  char *list[MAX_BREAKS];

  DebugBreakFillList(list);

  ret = IupListDialog(1, "Remove Break", debug.total_breaks, list, 0, 30, 10, NULL);

  for(i=0; i < debug.total_breaks; i++)
    free(list[i]);

  if (ret == -1)
    return IUP_DEFAULT;

  DebugRemoveBreak(ret);

  return IUP_DEFAULT;
}

static int DebugAddBreakDlg(void)
{
  char filename[1024] = "*.lua";
  if (!utlGetLuaFile(filename, 0))
    return IUP_DEFAULT;

  {
    int line = 1;

    char *fmt =
    {
      "Add Break\n"
      "line = %5.5%d\n"
    };

    if (IupScanf(fmt, &line) != 1)
      return IUP_DEFAULT;

    {
      int num_lin;
      char* text = utlLoadTextFile(filename);
      if (!text)
      {
        IupMessage("Error!", "Invalid filename.");
        return IUP_DEFAULT;
      }

      num_lin = utlStrLines(text);
      free(text);

      if (line < 1 || line > num_lin)
      {
        IupMessage("Error!", "Invalid line number."); 
        return IUP_DEFAULT;
      }
    }

    {
      luaDebugBreak* break_line = &debug.breaks[line];
      break_line->filename[break_line->count] = utlStrDup(filename);
      break_line->count++;
      debug.total_breaks++;

      if (strcmp(debug.filename, filename) == 0)
        DebugUpdateSourceBreak(line);
    }
  }

  return IUP_DEFAULT;
}

static void DebugKillBreaks(void)
{
  int i,j, n=0;
  for(i=0; i < MAX_BREAKS; i++)
  {
    luaDebugBreak* break_line = &debug.breaks[i];

    for (j=0; j < break_line->count; j++)
    {
      free(break_line->filename[j]);
      n++;

      if (n == debug.total_breaks)
      {
        debug.total_breaks = 0;
        break_line->count = 0;
        return;
      }
    }

    break_line->count = 0;
  }
}

static int DebugRemoveAllBreaks(void)
{
  DebugKillBreaks();

  {
    char* text = IupGetAttribute(debug.mlSource, IUP_VALUE);
    char *line_ptr;
    int line=1, changed=0, size;

    if (!text || text[0]==0)
      return IUP_DEFAULT;

    text = utlStrDup(text);
    size = strlen(text);

    do
    {
      if (line == 1)
        line_ptr = text + 4;
      else
        line_ptr = strstr(line_ptr, "\n") + 5;

      if (!line_ptr || line_ptr - text > size)
        break;

      if (*line_ptr == 'B')
      {
        *line_ptr = ' ';
        changed = 1;
      }

      line++;
    } while (line_ptr - text < size);

    if(changed)
      IupStoreAttribute(debug.mlSource, IUP_VALUE, text);

    free(text);
  }

  return IUP_DEFAULT;
}

static int DebugBreakpoints(void)
{
  Ihandle* menu;
  char *has_break = "ACTIVE=NO";

  if (debug.total_breaks > 0)
    has_break = "ACTIVE=YES";

  menu = IupMenu(
    IupItem ("Add File...", "luaDebugAddBreakDlg"),
    IupSetAttributes(IupItem ("Remove...", "luaDebugRemoveBreakDlg"), has_break),
    IupSetAttributes(IupItem ("Remove All", "luaDebugRemoveAllBreaks"), has_break),
    NULL);

  IupPopup(menu,IUP_MOUSEPOS,IUP_MOUSEPOS);

  return IUP_DEFAULT;
}

static int DebugFindFileNameBreak(char* filename, int line)
{
  luaDebugBreak* break_line = &debug.breaks[line];
  int i;

  for (i=0; i < break_line->count; i++)
  {
    if (strcmp(break_line->filename[i], filename) == 0)
      return i;
  }

  return -1;
}

static void DebugToggleBreak(int line)
{
  int break_file = DebugFindFileNameBreak(debug.filename, line);
  luaDebugBreak* break_line = &debug.breaks[line];

  if (break_file != -1)
  {
    free(break_line->filename[break_file]);

    if (break_line->count == 1)
      break_line->filename[break_file] = NULL;
    else
      break_line->filename[break_file] = break_line->filename[break_line->count-1];

    break_line->count--;
    debug.total_breaks--;
  }
  else
  {
    break_line->filename[break_line->count] = utlStrDup(debug.filename);
    break_line->count++;
    debug.total_breaks++;
  }
}

static int DebugHasLineBreak(int line)
{
  if (line <= 0)
    return 0;

  if (DebugFindFileNameBreak(debug.filename, line) != -1)
    return 1;

  return 0;
}

static void DebugUpdateSourceBreak(int line)
{
  /* changes the text displayed to contains the "B" mark at a line break */
  char* text = utlStrDup(IupGetAttribute(debug.mlSource, IUP_VALUE));
  char *line_ptr;

  if (line == 1)
    line_ptr = text + 4;
  else
  {
    char buffer[10];
    sprintf(buffer, "\n%03d:", line);
    line_ptr = strstr(text, buffer) + 5;
  }

  if (line_ptr)
  {
    if (DebugHasLineBreak(line))
      *line_ptr = 'B';
    else
      *line_ptr = ' ';

    IupStoreAttribute(debug.mlSource, IUP_VALUE, text);
  }

  free(text);
}

static int DebugBreakAtCaretLine(void)
{
  /* called when the user press the button "B" */

  int line = 1;
  sscanf(IupGetAttribute(debug.mlSource, IUP_CARET), "%d", &line);

  DebugToggleBreak(line);
  DebugUpdateSourceBreak(line);

  return IUP_DEFAULT;
}


/*************************************************************************/
/*                         Source                                        */
/*************************************************************************/

/* DebugUpdateSourceLine -> DebugUpdateSourceFunc -> DebugUpdateSourceText */

static void DebugUpdateSourceText(char* filename, char *text, int isfile)
{
  /* it is only called from DebugUpdateSourceFunc or 
     to clean up the source multiline */
  char* buffer, *title, *next_line, *line, b;
  int numlines, c, l, size;

  IupSetAttribute(debug.btCurLineBreak, IUP_ACTIVE, IUP_YES);

  debug.last_filename = filename;
  free(debug.filename);

  if (isfile)
    debug.filename = utlFullPath(filename);
  else
    debug.filename = utlStrDup(filename);

  title = utlStrDupUntil(filename, '\n');
  IupSetfAttribute(debug.lbSource, IUP_TITLE, "[ %s ]", title);
  free(title);

  if (!text) /* just update the title */
    return;

  if (text[0] == 0) /* clean up */
  {
    IupSetAttribute(debug.mlSource, IUP_VALUE, "");
    return;
  }

  numlines = utlStrLines(text);
  buffer = malloc(strlen(text) + numlines*8 + 1);

  c=0; l=1;
  line = text;

  do
  {
    next_line = utlGetLine(line, &size);

    if (DebugHasLineBreak(l))
      b = 'B';
    else
      b = ' ';

    c+=sprintf(buffer+c, "%03d:%c| ", l, (int)b);
    strncpy(buffer+c, line, size);
    c+=size; buffer[c] = '\n';
    c++; buffer[c] = '\0';

    l++;
    line = next_line;

  } while(*next_line != 0);

  IupStoreAttribute(debug.mlSource, IUP_VALUE, buffer);

  free(buffer);
}

static void DebugUpdateSourceFunc(char *filename, int linedefined)
{
  /* updates the source with the right buffer */
  /* checks if do not need to update again */
  /* called only by DebugUpdateSourceLine */
  /* should not be called for C functions, but just in case we check */
  if (linedefined == LINEDEF_CFUNC)
  {
    DebugUpdateSourceText("<cfunction>", NULL, 0);
  }
  else                   
  {
    /* main code of a chunk or function definition */  

    if (filename[0] != '@')
    {
      /* remove the " " from start and from end of the string */
      char *text = utlStrDup(filename);
      DebugUpdateSourceText(filename, text, 0);
      free(text);
    }
    else
    {
      if (strcmp(debug.last_filename, filename+1) != 0)
      {
        char *text = utlLoadTextFile(filename+1);
        if (!text)
        {
          IupMessage("Error!", "Can not load file.");
          return;
        }

        DebugUpdateSourceText(filename+1, text, 1);
        free(text);
      }
    }
  }
}

static void DebugUpdateSourceLine(int line)
{
  /* ignore not defined lines */
  if (line <= 0)
    return;

  /* updates the source at the current line */
  /* sets the cursor to the current line */
  /* called only when PAUSED or Animation or explicitly to show up the current line */
  /* it is called only in lua functions at the line hook */
  {
    lua_Function func;
    char *filename;
    int linedefined;

    func = lua_stackedfunction(0);
    lua_funcinfo(func, &filename, &linedefined);

    DebugUpdateSourceFunc(filename, linedefined);
  }

  IupSetfAttribute(debug.mlSource, IUP_CARET, "%d,7", line);
  IupSetfAttribute(debug.mlSource, IUP_SELECTION, "%d,1:%d,7", line, line);
  IupSetFocus(debug.mlSource);
  IupLoopStep();
}

static int DebugCurLine(void)
{
  /* force the display of the current line */

  int line = lua_currentline(lua_stackedfunction(0));
  if (line > 0)
    DebugUpdateSourceLine(line);
  return IUP_DEFAULT;
}


/*************************************************************************/
/*                      Local Watch                                      */
/*************************************************************************/

static char* DebugGetObject(lua_Object o)
{
  /* the local variables display shoes only strings or numbers
     other variables are only represented by its type */

  if (lua_rawisstring(o))
    return lua_getstring(o);
  else if (lua_isnumber(o))
  {
    static char buf[50];
    double number = lua_getnumber(o);

    if (number == floor(number))
      sprintf(buf, "%d", (int)number);
    else
      sprintf(buf, "%f", number);

    return buf;
  }
  else if (lua_istable(o))
    return "<table>";
  else if (lua_iscfunction(o))
    return "<cfunction>";
  else if (lua_isfunction(o))
    return "<function>";
  else if (lua_isuserdata(o))
    return "<userdata>";
  else if (lua_isnil(o))
    return "<nil>";
  else
    return "<unknown>";
}

static int DebugPrintLocal(void)
{
  lua_Object o, func;
  int local_number, level;
  char *name;

  /* prints the value of a local variable in the console */

  level = IupGetInt(debug.lsStack, IUP_VALUE)-1;
  level = level<0? 0: level;

  local_number = IupGetInt(debug.lsLocal, IUP_VALUE);
  local_number = local_number<1? 1: local_number;

  /* local variables always have a name */
  func = lua_stackedfunction(level);
  o = lua_getlocal(func, local_number, &name);

  luaConsolePrintObject(name, o);

  return IUP_DEFAULT;
}

static void lua_error_func(void);
static void dummy_error_func(void){}

static int DebugSetLocal(void)
{
  lua_Object o, func;
  int local_number, level;
  char *name, format[100], expr[100];

  /* change the value of a local variable */

  level = IupGetInt(debug.lsStack, IUP_VALUE)-1;
  level = level<0? 0: level;

  local_number = IupGetInt(debug.lsLocal, IUP_VALUE);
  local_number = local_number<1? 1: local_number;

  func = lua_stackedfunction(level);
  o = lua_getlocal(func, local_number, &name);

  sprintf(format, "Set Local\n%s = %%100.20%%s\n", name);

  if (lua_rawisstring(o))
    strcpy(expr, lua_getstring(o));
  else if (lua_isnumber(o))
  {
    double number = lua_getnumber(o);

    if (number == floor(number))
      sprintf(expr, "%d", (int)number);
    else
      sprintf(expr, "%f", number);
  }

  if (IupScanf(format, expr) != 1)
    return IUP_DEFAULT;

  lua_pushcfunction(dummy_error_func);
  lua_seterrormethod();

  /* check if the user enters an expression */

  sprintf(format, "return %s", expr);
  if (lua_dostring(format)==0)
  {
    o = lua_getresult(1);
    lua_pushobject(o);
    if (lua_setlocal(func, local_number)==1)
    {
      sprintf(format, "%d", local_number);
      IupSetfAttribute(debug.lsLocal, format, "%s = %s", name, DebugGetObject(o));
      IupStoreAttribute(debug.lsLocal, IUP_VALUE, format);
    }
    else
      IupMessage("Error!", "Can not set local variable.");
  }
  else
    IupMessage("Error!", "Invalid Expression.");

  lua_pushcfunction(lua_error_func);
  lua_seterrormethod();

  return IUP_DEFAULT;
}

static void DebugClearLocal(void)
{
  IupSetAttribute(debug.btPrintLocal, IUP_ACTIVE, IUP_NO);
  IupSetAttribute(debug.btSetLocal, IUP_ACTIVE, IUP_NO);
  IupSetAttribute(debug.lsLocal, "1", NULL);
}

static void DebugUpdateLocal(lua_Object func)
{
  lua_Object o;
  int local_number = 1;
  char *name, buf[10];

  /* updates the local variables display */

  lua_beginblock();

  while((o = lua_getlocal(func, local_number, &name)) != LUA_NOOBJECT)
  {
    lua_beginblock();

    sprintf(buf, "%d", local_number);
    IupSetfAttribute(debug.lsLocal, buf, "%s = %s", name, DebugGetObject(o));

    local_number++;

    lua_endblock();
  }

  lua_endblock();

  sprintf(buf, "%d", local_number);
  IupSetAttribute(debug.lsLocal, buf, NULL);

  if (local_number == 1)
  {
    IupSetAttribute(debug.btPrintLocal, IUP_ACTIVE, IUP_NO);
    IupSetAttribute(debug.btSetLocal, IUP_ACTIVE, IUP_NO);
  }
  else
  {
    IupSetAttribute(debug.btPrintLocal, IUP_ACTIVE, IUP_YES);
    IupSetAttribute(debug.btSetLocal, IUP_ACTIVE, IUP_YES);
    IupSetAttribute(debug.lsLocal, IUP_VALUE, "1");
  }
}


/*************************************************************************/
/*                      Call Stack                                       */
/*************************************************************************/


static int DebugFuncInfoStr(lua_Object func, char* buf)
{
  char *name;
  char *filename = NULL;
  int linedefined, currentline;
  int c=0;

  /* guess the function name, type,
     the current line where it is executing and where it was defined */

  lua_funcinfo(func, &filename, &linedefined);

  switch (*lua_getobjname(func, &name)) 
  {
  case 'g': /* global */
    if (linedefined == LINEDEF_CFUNC)
      c+=sprintf(buf+c, "%s (C)", name);
    else
      c+=sprintf(buf+c, "%s", name);
    break;
  case 't': /* tag-method */
    if (linedefined == LINEDEF_CFUNC)
      c+=sprintf(buf+c, "\"%s\" tag method (C)", name);
    else
      c+=sprintf(buf+c, "\"%s\" tag method", name);
    break;
  default: 
    {
      if (linedefined == LINEDEF_MAIN)       
        c+=sprintf(buf+c, "<main>");
      else if (linedefined == LINEDEF_CFUNC)   
        c+=sprintf(buf+c, "<no name> (C)");
      else                        
        c+=sprintf(buf+c, "<no name>"); /* Lua function */
    }
  }

  currentline = lua_currentline(func);

  if (currentline > 0)
    c+=sprintf(buf+c, ", at line %d\n", currentline);
  else
    c+=sprintf(buf+c, "\n");

  if (linedefined != LINEDEF_CFUNC)
  {
    if (linedefined == LINEDEF_MAIN)
    {
      if (filename[0] != '@')
      {
        char* text = utlStrDupUntil(filename, '\n');

        if (text[0] != 0)
          c+=sprintf(buf+c, "    [defined in the string \"%s\"...]\n", text);
        else
          c+=sprintf(buf+c, "    [defined in a string]\n");

        free(text);
      }
      else
      {
        char* text = utlGetFileTitle(filename+1);
        c+=sprintf(buf+c, "    [defined in the file \"%s\"]\n", text);
        free(text);
      }
    }
    else
    {
      if (filename[0] != '@')
        c+=sprintf(buf+c, "    [defined in a string at line %d]\n", linedefined);
      else
      {
        char* text = utlGetFileTitle(filename+1);
        c+=sprintf(buf+c, "    [defined in the file \"%s\" at line %d]\n", text, linedefined);
        free(text);
      }
    }
  }

  return c;
}

static void PrintStack(int level)
{
  char buf[10240];
  lua_Object func;
  int c = 0, start_level = level;

  /* prints the stack in the console, 
     here we print more information than in the stack display */

  c+=sprintf(buf+c, "Stack:\n");

  while ((func = lua_stackedfunction(level)) != LUA_NOOBJECT) 
  {
    c+=sprintf(buf+c, "% 3d| ", level);
    c+=DebugFuncInfoStr(func, buf+c);
    level++;
  }

  if (start_level == level)
    c+=sprintf(buf+c, "  (empty)");

  luaConsoleEnterMessage(buf);
}

static int DebugPrintStack(void)
{
  PrintStack(0);
  return IUP_DEFAULT;
}

static int DebugPrintLevel(void)
{
  int level;
  lua_Object func;
  int c=0;
  char buf[200];

  /* called when the user press the button print and
     a function is selected at the stack display */

  level = IupGetInt(debug.lsStack, IUP_VALUE)-1;
  level = level<0? 0: level;
  func = lua_stackedfunction(level);

  c+=sprintf(buf+c, "Stack Level %d:\n", level);
  c+=DebugFuncInfoStr(func, buf+c);
  luaConsoleEnterMessage(buf);

  return IUP_DEFAULT;
}

static int DebugLevel(void)
{
  int level = -1;
  lua_Object func;

  /* calculates the curret debug level in the call stack */

  do
  {
    level++;
    func = lua_stackedfunction(level);
  } while (func != LUA_NOOBJECT);

  return level-1;
}

static void DebugClearStack(void)
{
  DebugClearLocal();

  debug.last_stack_update_level = -1;
  IupSetAttribute(debug.btPrintLevel, IUP_ACTIVE, IUP_NO);
  IupSetAttribute(debug.btPrintStack, IUP_ACTIVE, IUP_NO);
  IupSetAttribute(debug.lsStack, "1", NULL);
}

static int DebugSelectFunc(Ihandle *self, char *t, int i, int v)
{
  /* called when the user selects a function in the stack list display */
  if (v == 1)
    DebugUpdateLocal(lua_stackedfunction(i-1));

  return IUP_DEFAULT;
}

static void DebugUpdateStack(void)
{
  int level = 0;  
  lua_Object func;
  char buf[10];

  /* updates the stack list display */

  if (debug.last_stack_update_level == DebugLevel())
  {
    level = IupGetInt(debug.lsStack, IUP_VALUE)-1;
    level = level<0? 0: level;
    DebugUpdateLocal(lua_stackedfunction(level));
    return;
  }

  while ((func = lua_stackedfunction(level++)) != LUA_NOOBJECT) 
  {
    char *filename;
    int linedefined;
    char *name;

    lua_beginblock();

    if (*lua_getobjname(func, &name) == 0) 
    {
      lua_funcinfo(func, &filename, &linedefined);
      if (linedefined == LINEDEF_MAIN)        
        name = "<main>";
      else
        name = "<noname>";
    }

    sprintf(buf, "%d", level);
    IupStoreAttribute(debug.lsStack, buf, name);

    lua_endblock();
  }

  sprintf(buf, "%d", level);
  IupSetAttribute(debug.lsStack, buf, NULL);

  if (level > 1)
  {
    IupSetAttribute(debug.btPrintLevel, IUP_ACTIVE, IUP_YES);
    IupSetAttribute(debug.btPrintStack, IUP_ACTIVE, IUP_YES);
    IupSetAttribute(debug.lsStack, IUP_VALUE, "1");

    DebugUpdateLocal(lua_stackedfunction(0));
  }
  else
  {
    IupSetAttribute(debug.btPrintLevel, IUP_ACTIVE, IUP_NO);
    IupSetAttribute(debug.btPrintStack, IUP_ACTIVE, IUP_NO);
    DebugClearLocal();
  }

  debug.last_stack_update_level = level-2;
}


/*************************************************************************/
/*                         Hooks                                         */
/*************************************************************************/


static void DebugSetState(int state);
static void DebugLineHook(int line);

static void DebugCallHook(lua_Function func, char *filename, int linedefined)
{
  int level = DebugLevel();

  if (func == LUA_NOOBJECT)
  {
    /* getting out of a function */

    /* checks if I'm in a middle of a STOP */
    if (debug.levels_to_stop != 0)
    {
      debug.levels_to_stop = level+1;
      lua_error(NULL); /* this will make a long jump */
    }

    /* checks if I'm getting OUT from the base MAIN */
    if (level == 0)
    {
      DebugSetState(DEBUG_ACTIVE);
      return;
    }

    if (debug.step_mode == MODE_OUT || 
        (debug.step_mode == MODE_LINES && debug.step_func_mode == STEP_OVER))
    {
      /* checks if I'm getting out from the function I get in */
      if (debug.step_func_level == level)
      {
        debug.step_func_state = FUNC_OUTSIDE;
        debug.step_func_level = 0;
      }
    }
  }
  else
  {
    /* getting in the function 'func' */

    /* checks if I'm getting IN the base MAIN */
    if (level == 0)
    {
      /* this should happen only after a sucessfull STOP */
      if (debug.levels_to_stop != 0)
        debug.levels_to_stop = 0;

      /* always PAUSED in the begging of MAIN */
      DebugSetState(DEBUG_PAUSED);

      debug.step_func_state = FUNC_OUTSIDE;
      debug.step_func_level = 0;
      return;
    }

    /* checks if I'm in a middle of a STOP */
    if (debug.levels_to_stop != 0)
    {
      debug.levels_to_stop = level+1;
      lua_error(NULL); /* this will make a long jump */
    }

    if (debug.step_mode == MODE_LINES && debug.step_func_mode == STEP_OVER)
    {
      /* if I'm in the first call of this mode (it can happen in any level of the stack) */
      if (debug.step_func_level == 0)
      {
        /* get in, mark and wait until gets out */
        debug.step_func_state = FUNC_INSIDE;
        debug.step_func_level = level;
      }
    }
  }
}

static void DebugUpdateState(int line)
{
  /* called only by the line hook */
  /* check STEP and breaks, turn to PAUSED if necessary */

  if (debug.state == DEBUG_STEP)
  {
    if (debug.step_mode == MODE_LINES)
    {
      if (debug.step_func_mode == STEP_INTO || 
          (debug.step_func_mode == STEP_OVER && debug.step_func_state == FUNC_OUTSIDE))
      {
        debug.step_lines_count++;
        if (debug.step_lines_count == debug.step_lines_total)
        {
          DebugSetState(DEBUG_PAUSED);
          debug.step_lines_count = 0;
        }
      }
    }
    else /* debug.step_mode == MODE_OUT */
    {
      /* if I just get out, then PAUSED */
      if (debug.step_func_state == FUNC_OUTSIDE)
      {
        int level = DebugLevel();

        DebugSetState(DEBUG_PAUSED);

        /* prepare to get out from where I am now */
        debug.step_func_state = FUNC_INSIDE;
        debug.step_func_level = level;
      }
    }
  }

  /* check breakpoits only if I'm not PAUSED already */
  if (debug.state != DEBUG_PAUSED)
  {
    char *filename;
    int linedefined;

    lua_funcinfo(lua_stackedfunction(0), &filename, &linedefined);
    free(debug.filename);
    if (filename[0] != '@')
      debug.filename = utlStrDup(filename);
    else
      debug.filename = utlFullPath(filename+1);

    if (DebugHasLineBreak(line))
      DebugSetState(DEBUG_PAUSED);
  }
}

static int DebugWaitIdle(void)
{
  if (IupLoopStep() == IUP_CLOSE)
  {
    debug.state = DEBUG_STOPPED;
    return 1;
  }

  return 0;
}

static void DebugLineHook(int line)
{
  /* checks if I'm in a middle of a STOP */
  if (debug.levels_to_stop != 0)
  {
    debug.levels_to_stop = DebugLevel()+1;
    lua_error(NULL); /* this will make a long jump */
  }

  /* check STEP and breaks, turn to PAUSED if necessary */
  DebugUpdateState(line);

  /* animation is done only when RUN/STEP (step at this point means a long step Over/Out/N>1) */
  if ((debug.state == DEBUG_RUN || debug.state == DEBUG_STEP) && debug.wait_time != 0)
  {
    DebugUpdateSourceLine(line);
    DebugUpdateStack();
    utlWait(debug.wait_time, DebugWaitIdle);
  }

  if (debug.state == DEBUG_PAUSED)
  {
    DebugUpdateSourceLine(line);
    DebugUpdateStack();

    /* disable debugging, so the user can consult variables and executes Lua code freely */
    lua_setlinehook(NULL);
    lua_setcallhook(NULL);

    /* waits for the user to press: Stop/Play/Step */
    while (debug.state == DEBUG_PAUSED)
    {
      if (IupLoopStep() == IUP_CLOSE)
        debug.state = DEBUG_STOPPED;
    }

    lua_setlinehook(DebugLineHook);
    lua_setcallhook(DebugCallHook);
  }
  else
  {
    if (IupLoopStep() == IUP_CLOSE)
      debug.state = DEBUG_STOPPED;
  }

  /* if it was RUN/STEP/PAUSED and I pressed STOP */
  if (debug.state == DEBUG_STOPPED)
  {
    luaConsoleEnterMessage("debug: Last command interrupted!");
    DebugSetState(DEBUG_ACTIVE);
    debug.levels_to_stop = DebugLevel()+1;
    lua_error(NULL); /* this will make a long jump */
  }
}


/*************************************************************************/
/*                         Control                                       */
/*************************************************************************/


static void DebugSetState(int state)
{
  char *title, *stop, *step, *pause, *play, *curline;

  switch(state)
  {
  case DEBUG_INACTIVE:
    title = "Lua Debug (Inactive)";
    stop = IUP_NO;
    step = IUP_NO;
    play = IUP_NO;
    pause = IUP_NO;
    curline = IUP_NO;

    DebugUpdateSourceText("", "", 0);
    DebugClearStack();
    lua_setlinehook(NULL);
    lua_setcallhook(NULL);
    IupSetAttribute(debug.tgActive, IUP_VALUE, IUP_OFF);
    break;
  case DEBUG_ACTIVE:
    title = "Lua Debug (Active)";
    stop = IUP_NO;
    step = IUP_NO;
    play = IUP_NO;
    pause = IUP_NO;
    curline = IUP_NO;

    DebugUpdateSourceText("", "", 0);
    DebugClearStack();
    lua_setlinehook(DebugLineHook);
    lua_setcallhook(DebugCallHook);
    IupSetAttribute(debug.tgActive, IUP_VALUE, IUP_ON);
    break;
  case DEBUG_RUN:
  case DEBUG_STEP:
    title = "Lua Debug (Running)";
    stop = IUP_YES;
    step = IUP_NO;
    play = IUP_NO;
    pause = IUP_YES;
    curline = IUP_NO;

    debug.step_lines_count = 0;

    if (debug.step_mode == MODE_LINES)
    {
      debug.step_func_state = FUNC_OUTSIDE;
      debug.step_func_level = 0;
    }
    else
    {
      int level = DebugLevel();
      debug.step_func_state = FUNC_INSIDE;
      debug.step_func_level = level;
    }

    break;
  case DEBUG_PAUSED:
    title = "Lua Debug (Paused)";
    stop = IUP_YES;
    step = IUP_YES;
    play = IUP_YES;
    pause = IUP_NO;
    curline = IUP_YES;
    break;
  }

  debug.state = state;
  IupSetAttribute(debug.iupDialog, IUP_TITLE, title);
  IupSetAttribute(debug.btStop, IUP_ACTIVE, stop);
  IupSetAttribute(debug.btStep, IUP_ACTIVE, step);
  IupSetAttribute(debug.btPause, IUP_ACTIVE, pause);
  IupSetAttribute(debug.btPlay, IUP_ACTIVE, play);
  IupSetAttribute(debug.btCurLine, IUP_ACTIVE, curline);
  IupSetAttribute(debug.btCurLineBreak, IUP_ACTIVE, curline);
}

static int DebugAnimateDelayDlg(void)
{
  int wait = debug.wait_time;

  char *fmt =
  {
    "Animate Delay\n"
    "time (milisec.) = %20.10%d\n"
  };

  if (IupScanf(fmt, &wait) != 1)
    return IUP_DEFAULT;

  debug.wait_time = wait;

  return IUP_DEFAULT;
}

static int DebugStepLinesNum(Ihandle* self, int c, char *after)
{
  int step_lines = 0;
  sscanf(after, "%d", &step_lines);

  if (step_lines <= 0 || debug.step_lines_total == step_lines)
    return IUP_DEFAULT;

  debug.step_lines_total = step_lines;
  debug.step_lines_count = 0;

  return IUP_DEFAULT;
}

static int DebugStepLines(void)
{
  if (debug.step_mode == MODE_LINES)
    return IUP_DEFAULT;

  debug.step_mode = MODE_LINES;
  debug.step_lines_count = 0;

  debug.step_func_state = FUNC_OUTSIDE;
  debug.step_func_level = 0;

  lua_setlinehook(DebugLineHook);

  return IUP_DEFAULT;
}

static int DebugStepOut(void)
{
  int level;

  if (debug.step_mode == MODE_OUT)
    return IUP_DEFAULT;

  level = DebugLevel();

  debug.step_mode = MODE_OUT;
  debug.step_lines_count = 0;

  debug.step_func_state = FUNC_INSIDE;
  debug.step_func_level = level;

  return IUP_DEFAULT;
}

static int DebugStepInto(void)
{
  debug.step_func_mode = STEP_INTO;
  return IUP_DEFAULT;
}

static int DebugStepOver(void)
{
  debug.step_func_mode = STEP_OVER;
  return IUP_DEFAULT;
}

static int DebugActive(Ihandle* self, int v)
{
  if (v == 1)
    DebugSetState(DEBUG_ACTIVE);
  else
    DebugSetState(DEBUG_INACTIVE);

  return IUP_DEFAULT;
}

static int DebugStop(void)
{
  debug.state = DEBUG_STOPPED;
  return IUP_DEFAULT;
}

static int DebugStep(void)
{
  DebugSetState(DEBUG_STEP);
  return IUP_DEFAULT;
}

static int DebugPlay(void)
{
  DebugSetState(DEBUG_RUN);
  return IUP_DEFAULT;
}

static int DebugPause(void)
{
  DebugSetState(DEBUG_PAUSED);
  return IUP_DEFAULT;
}

static int DebugWillAbort(void)
{
  int level = 0;  
  lua_Object func;
  int has_main = 0; 

  /* guess if the error in lua_error_func will stop the debuging process */

  while ((func = lua_stackedfunction(level++)) != LUA_NOOBJECT) 
  {
    char *filename, *name;
    int linedefined;

    if (has_main)         /* there is a function call in a level deeper than the main founded */
      return 0;           /* the error will NOT finish the debugging procedure */

    lua_funcinfo(func, &filename, &linedefined);
    if (linedefined == LINEDEF_MAIN) 
      has_main = 1;       /* found a main */

    lua_getobjname(func, &name);
    if (name && (strcmp(name, "dostring") == 0 || strcmp(name, "dofile") == 0))
      has_main = 1;       /* found a pre-main, this will catch compile time errors */
  }

                          /* there is no main, or the main is at the base */
  return 1;               /* the error will finish the debugging procedure */
}

static void lua_error_func(void)
{
  if (DebugWillAbort() && debug.state != DEBUG_INACTIVE)      
    DebugSetState(DEBUG_ACTIVE);  /* prepare to finish the debugging procedure */

  debug.old_error_func();

  if (lua_stackedfunction(1) != LUA_NOOBJECT) 
    PrintStack(1); /* skip level 0 (it's the error function) */
}


/*************************************************************************/
/*                         Window                                        */
/*************************************************************************/


int luaDebugShow(void)
{
  IupShow(debug.iupDialog);
  lua_setdebug(1);
  return IUP_DEFAULT;
}

int luaDebugHide(void)
{
  IupHide(debug.iupDialog);
  lua_setdebug(0);
  return IUP_DEFAULT;
}

static int DebugClose(void)
{
  return debug.close;
}

void luaDebugKill(void)
{
  free(debug.filename);
  DebugKillBreaks();
  lua_setdebug(0);
  IupDestroy(debug.iupDialog);
}

void luaDebugCreate(char* parent, int close)
{
  Ihandle *vbControls, *hbToolBar, *hbCtrlBar, *frStepMode, *frStepFunc, 
          *frSource, *frStack, *frLocal, *frCtrl;

  /* everything is initialized with zeros */
  memset(&debug, 0, sizeof(struct _luaDebug));

  debug.close = close;
  debug.step_lines_total = 1;
  debug.last_filename = "";
  debug.filename = utlStrDup("");

  IupSetFunction("luaDebugClose", (Icallback)DebugClose);
  IupSetFunction("luaDebugPrintLevel", (Icallback)DebugPrintLevel);
  IupSetFunction("luaDebugPrintStack", (Icallback)DebugPrintStack);
  IupSetFunction("luaDebugBreakAtCaretLine", (Icallback)DebugBreakAtCaretLine);
  IupSetFunction("luaDebugCurLine", (Icallback)DebugCurLine);
  IupSetFunction("luaDebugActive", (Icallback)DebugActive);
  IupSetFunction("luaDebugAnimateDelayDlg", (Icallback)DebugAnimateDelayDlg);
  IupSetFunction("luaDebugStop", (Icallback)DebugStop);
  IupSetFunction("luaDebugStep", (Icallback)DebugStep);
  IupSetFunction("luaDebugPlay", (Icallback)DebugPlay);
  IupSetFunction("luaDebugPause", (Icallback)DebugPause);
  IupSetFunction("luaDebugStepLinesNum", (Icallback)DebugStepLinesNum);
  IupSetFunction("luaDebugStepLines", (Icallback)DebugStepLines);
  IupSetFunction("luaDebugStepOut", (Icallback)DebugStepOut);
  IupSetFunction("luaDebugStepInto", (Icallback)DebugStepInto);
  IupSetFunction("luaDebugStepOver", (Icallback)DebugStepOver);
  IupSetFunction("luaDebugSelectFunc", (Icallback)DebugSelectFunc);
  IupSetFunction("luaDebugPrintLocal", (Icallback)DebugPrintLocal);
  IupSetFunction("luaDebugSetLocal", (Icallback)DebugSetLocal);
  IupSetFunction("luaDebugRemoveAllBreaks", (Icallback)DebugRemoveAllBreaks);
  IupSetFunction("luaDebugRemoveBreakDlg", (Icallback)DebugRemoveBreakDlg);
  IupSetFunction("luaDebugAddBreakDlg", (Icallback)DebugAddBreakDlg);
  IupSetFunction("luaDebugBreakpoints", (Icallback)DebugBreakpoints);

  hbToolBar = IupSetAttributes(IupHbox(
    IupSetAttributes(IupLabel(""),"IMAGE=LuaButton"),
    IupSetAttributes(IupFill(), "EXPAND=NO, SIZE=10"),
    debug.tgActive = IupToggle("Active", "luaDebugActive"),
    NULL),"ALIGNMENT=ACENTER");;

  hbCtrlBar = IupSetAttributes(IupHbox(
    IupSetAttributes(debug.btStop = IupButton("", "luaDebugStop"), "TIP=\"Stops the execution.\", IMAGE=utlStopButton, IMINACTIVE=utlStopButtonGray, ACTIVE=NO"),
    IupSetAttributes(debug.btPause = IupButton("", "luaDebugPause"), "TIP=\"Pauses the execution.\", IMAGE=utlPauseButton, IMINACTIVE=utlPauseButtonGray, ACTIVE=NO"),
    IupSetAttributes(debug.btPlay = IupButton("", "luaDebugPlay"), "TIP=\"Continues the execution.\", IMAGE=utlPlayButton, IMINACTIVE=utlPlayButtonGray, ACTIVE=NO"),
    IupSetAttributes(debug.btStep = IupButton("", "luaDebugStep"), "TIP=\"Executes one or more steps of the execution.\", IMAGE=utlStepButton, IMINACTIVE=utlStepButtonGray, ACTIVE=NO"),
    NULL),"GAP=0");

  frCtrl = IupSetAttributes(IupFrame(IupVbox(
    IupFill(), 
    hbCtrlBar, 
    IupFill(), 
    NULL)), "TITLE=\"Control:\"");

  frStepMode = IupSetAttributes(IupFrame(IupRadio(IupVbox(
    IupFill(),
    IupSetAttributes(IupHbox(
      IupToggle("Lines:", "luaDebugStepLines"), 
      IupSetAttributes(IupText("luaDebugStepLinesNum"), "SIZE=40, VALUE=1, MASK=#####"), 
      NULL), "MARGIN=0x0, ALIGNMENT=ACENTER"),
    IupToggle("Out", "luaDebugStepOut"),
    IupFill(), 
    NULL))), "MARGIN=0x0, TITLE=\"Step Mode:\"");

  frStepFunc = IupSetAttributes(IupFrame(IupRadio(IupVbox(
    IupFill(),
    IupToggle("Into", "luaDebugStepInto"),
    IupToggle("Over", "luaDebugStepOver"),
    IupFill(), 
    NULL))), "SIZE=70x, GAP=10, TITLE=\"Step Function:\"");

  frSource = IupSetAttributes(IupFrame(IupVbox(
    IupSetAttributes(IupHbox(
      debug.lbSource = IupSetAttributes(IupLabel("[  ]"),"EXPAND=HORIZONTAL"), 
      debug.btCurLine = IupSetAttributes(IupButton("->", "luaDebugCurLine"),"FONT=COURIER_BOLD_10, SIZE=12x15, TIP=\"Shows the current line. Ctrl+L\", ACTIVE=NO"), 
      debug.btCurLineBreak = IupSetAttributes(IupButton("B", "luaDebugBreakAtCaretLine"),"FONT=COURIER_BOLD_10, SIZE=12x15, TIP=\"Toggle breakpoint at the current line. Ctrl+B\", ACTIVE=NO"), 
      NULL), "ALIGNMENT=ACENTER"), 
    debug.mlSource = IupSetAttributes(IupMultiLine("do_nothing"), "EXPAND=YES, BGCOLOR=\"224 224 224\", FONT=COURIER_NORMAL_12"), 
    NULL)), "TITLE=Source:");

  frStack = IupSetAttributes(IupFrame(IupHbox(
    debug.lsStack = IupSetAttributes(IupList("luaDebugSelectFunc"), "SIZE=50x, EXPAND=VERTICAL, FONT=COURIER_NORMAL_12"), 
    IupVbox(
      debug.btPrintLevel = IupSetAttributes(IupButton("Print", "luaDebugPrintLevel"),"TIP=\"Prints information about the selected call stack level.\", ACTIVE=NO, SIZE=35x"), 
      debug.btPrintStack = IupSetAttributes(IupButton("Print All", "luaDebugPrintStack"),"TIP=\"Prints information about all the stack levels.\", ACTIVE=NO, SIZE=35x"), 
      NULL), 
    NULL)), "TITLE=\"Call Stack:\"");

  frLocal = IupSetAttributes(IupFrame(IupHbox(
    debug.lsLocal = IupSetAttributes(IupList("donothing"),"SIZE=50x, EXPAND=VERTICAL, FONT=COURIER_NORMAL_12"), 
    IupVbox(
      debug.btPrintLocal = IupSetAttributes(IupButton("Print", "luaDebugPrintLocal"),"ACTIVE=NO, SIZE=35x"), 
      debug.btSetLocal = IupSetAttributes(IupButton("Set...", "luaDebugSetLocal"),"ACTIVE=NO, SIZE=35x"), 
      NULL), 
    NULL)), "TITLE=\"Local Watch:\"");

  vbControls = IupVbox(
    IupSetAttributes(hbToolBar, "MARGIN=4x4, GAP=8"),
    IupSetAttributes(IupHbox(
      frCtrl,
      frStepMode,
      frStepFunc,
      IupSetAttributes(IupVbox(
        IupSetAttributes(IupButton("Animate Delay...", "luaDebugAnimateDelayDlg"), "TIP=\"Sets the time delay between each line hook.\", SIZE=70x"), 
        IupSetAttributes(IupFill(), "EXPAND=NO, SIZE=7"),
        IupSetAttributes(IupButton("Breakpoints >>", "luaDebugBreakpoints"), "TIP=\"Add/Remove breakpoints.\", SIZE=70x"), 
        NULL),"MARGIN=4x8, GAP=0"),
    NULL),"MARGIN=4x4, GAP=8"),
    IupSetAttributes(IupHbox(
      frSource,
      IupSetAttributes(IupVbox(
        IupSetAttributes(frStack,"MARGIN=4x4, GAP=8"),
        IupSetAttributes(frLocal,"MARGIN=4x4, GAP=8"),
        NULL),"MARGIN=0x0"),
      NULL),"MARGIN=4x4, GAP=8"),
    NULL);

  debug.iupDialog = IupDialog(vbControls);
  if (parent) IupSetAttribute(debug.iupDialog,IUP_PARENTDIALOG, parent);
  IupSetAttribute(debug.iupDialog,IUP_CLOSE_CB,"luaDebugClose");
  IupSetAttribute(debug.iupDialog,IUP_TITLE, "Lua Debug (Inactive)");
  IupSetAttribute(debug.iupDialog,IUP_K_cL, "luaDebugCurLine");
  IupSetAttribute(debug.iupDialog,IUP_K_cB, "luaDebugBreakAtCaretLine");
  IupSetAttribute(debug.iupDialog,IUP_SIZE, "400x250");
  IupSetAttribute(debug.iupDialog,IUP_ICON, "LuaIcon");
  IupSetAttribute(debug.mlSource,IUP_READONLY, IUP_YES);

  IupSetHandle("luaDebug", debug.iupDialog);

  lua_register("ldShow", (lua_CFunction)luaDebugShow);

  IupShowXY(debug.iupDialog, 150, 180);

  lua_pushcfunction(lua_error_func);
  debug.old_error_func = lua_getcfunction(lua_seterrormethod());
}

